"""
WebApp.
"""
import re
from urllib.parse import parse_qs
from typing import Tuple, List, Callable
import azure.functions as func
from . import logger


class HTTPError(Exception):
	pass


class NotFound(HTTPError):
	pass


class MethodNotAllowed(HTTPError):
	pass


class WebApp:

	NOT_FOUND = func.HttpResponse(status_code=404)
	METHOD_NOT_ALLOWED = func.HttpResponse(status_code=405)
	SERVER_ERROR = func.HttpResponse(status_code=500)

	MESSAGES = {
		200: "OK",
		201: "Created",
		204: "No Content",
		301: "Moved Permanently",
		302: "Found",
		304: "Not Modified",
		401: "Unauthorized",
		403: "Forbidden",
		404: "Not Found",
		405: "Method Not Allowed",
		500: "Internal Server Error",
		503: "Service Unavailable"
	}

	def __init__(self, init_params: List[Tuple[str, List[str], Callable, str]]):
		# (route, methods, function, param_name)
		self.init_params = init_params
		self.find_params_names_regex = re.compile(r"\{(.*?)\}")
		self.find_regex_constraint = re.compile(r"regex\((.*?)\)")
		self.rules = []
		self.build_regex()
		
	def build_regex(self):
		for tup in self.init_params:
			params_names = self.find_params_names_regex.findall(tup[0])
			regex = tup[0] + "$"
			for param in params_names:

				if ":" in param:
					name, constraint = param.split(":", 1)
					try:
						constraint = self.find_regex_constraint.findall(constraint)[0]
						regex = regex.replace("{"+param+"}", "({})".format(constraint))
						params_names[params_names.index(param)] = name
					except:
						logger.warning("Constraint not supported: {}".format(tup[0]))
						regex = regex.replace("{"+param+"}", "([^/]*?)")
				else:
					regex = regex.replace("{"+param+"}", "([^/]*?)")

			route_match_regex = re.compile(regex)
			self.rules.append((route_match_regex, params_names, tup))

	def select_handler(self, method: str, path: str):
		for rule in self.rules:
			result = rule[0].match(path)
			if result is not None:
				if method not in rule[2][1]:
					raise MethodNotAllowed()
				else:
					return rule[2], dict(zip(rule[1], result.groups()))
		raise NotFound()

	def __call__(self, environ, start_response):
		return self.wsgi(environ, start_response)

	def environ_to_request(self, environ: dict, route_params: dict) -> func.HttpRequest:
		# query parameters
		args = parse_qs(environ["QUERY_STRING"])
		query = {}
		for key, value in args.items():
			query[key] = ",".join(value)
		# headers
		headers = {}
		for key in environ:
			if key.startswith("HTTP_"):
				headers[key.replace("HTTP_", "").replace("_", "-").lower()] = environ[key]
		# content length
		try:
			request_body_size = int(environ.get('CONTENT_LENGTH', 0))
		except (ValueError):
			request_body_size = 0

		return func.HttpRequest(
			environ["REQUEST_METHOD"],
			environ["PATH_INFO"],
			headers=headers,
			route_params=route_params,
			params=query,
			body=environ['wsgi.input'].read(request_body_size)
		)

	def response_to_body(self, response: func.HttpResponse):
		return [response.get_body()]

	def get_status_message(self, status_code: int) -> str:
		try:
			return self.MESSAGES[status_code]
		except KeyError:
			return ""

	def process(self, environ: str) -> func.HttpResponse:
		try:
			handler, route_params = self.select_handler(
				environ["REQUEST_METHOD"].lower(),
				environ["PATH_INFO"]
			)
			request = self.environ_to_request(environ, route_params)
			return handler[2](**{handler[3]: request})
		except NotFound:
			return self.NOT_FOUND
		except MethodNotAllowed:
			return self.METHOD_NOT_ALLOWED
		except BaseException as e:
			logger.exception("")
			return self.SERVER_ERROR

	def wsgi(self, environ, start_response):
		response = self.process(environ)
		start_response(
			"{} {}".format(
				response.status_code,
				self.get_status_message(response.status_code)
			),
			list(response.headers.items())
		)
		return self.response_to_body(response)



"""
Module to look for, and parse, Function and settings.
"""
import os
import json
from . import logger


def is_function(folder: str) -> bool:
	return os.path.isfile("{}/function.json".format(folder))


def get_functions(path: str) -> list:
	functions = []
	for file in os.listdir(path):
		candidate = "{}/{}".format(path, file)
		if os.path.isdir(candidate):
			if is_function(candidate):
				functions.append((candidate, file))
	return functions


def load_json_file(path: str) -> dict:
	with open(path, "r") as file:
		return json.load(file)


def load_function_settings(path: str) -> dict:
	return load_json_file("{}/function.json".format(path))


def load_project_settings(path: str) -> dict:
	return load_json_file("{}/host.json".format(path))


def load_project(path: str) -> dict:
	try:
		project_settings = load_project_settings(path)
	except json.decoder.JSONDecodeError:
		logger.error("Unable to parse host.json: invalid JSON.")
		project_settings = {}
	except FileNotFoundError:
		logger.error("Unable to parse host.json: file not found.")
		project_settings = {}

	functions_settings = []
	functions = get_functions(path)
	for function in functions:
		try:
			functions_settings.append((function[1], load_function_settings(function[0])))
		except json.decoder.JSONDecodeError:
			logger.error("[{}] Unable to parse Function settings: invalid JSON.".format(function[1]))
	return {"project": project_settings, "functions": functions_settings}



"""
Function entities.
"""
from importlib import import_module


class Project:
	def __init__(self, settings: dict):
		self.settings = settings

	@property
	def route_prefix(self):
		prefix = self.settings.get("extensions", {}).get("http", {}).get("routePrefix", "/")
		if not prefix.startswith("/"):
			prefix = "/" + prefix
		return prefix


class Binding:

	def __init__(self, settings: dict):
		self.settings = settings

	@property
	def type(self):
		return self.settings.get("type")

	@property
	def direction(self):
		return self.settings.get("direction")

	@property
	def name(self):
		return self.settings.get("name")


class HTTP(Binding):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if self.direction != "out":
			raise ValueError("http binding direction must be out")
		if self.name != "$return":
			raise ValueError("http binding name must be $return")


class HTTPTrigger(Binding):

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if self.direction != "in":
			raise ValueError("httpTrigger binding direction must be in")
		if self.authLevel != "Anonymous":
			raise ValueError("httpTrigger binding authLevel must be Anonymous")

	@property
	def authLevel(self):
		return self.settings.get("authLevel", "Anonymous")

	@property
	def methods(self):
		return self.settings.get("methods", [])

	@property
	def route(self):
		return self.settings.get("route", [])


class Function:

	def __init__(self, func_name: str, settings: dict):
		self.settings = settings
		self.func_name = func_name
		temp = self.bindings
		if len(temp) > 2:
			raise ValueError("Maximum two bindings")
		a, b = tuple(temp)
		if b.type == "httpTrigger" and a.type == "http":
			t = a
			a = b
			b = t
		self.trigger = HTTPTrigger(a.settings)
		self.output = HTTP(b.settings)

	@property
	def bindings(self):
		return list(map(lambda x: Binding(x), self.settings.get("bindings", [])))
	
	@property
	def scriptFile(self):
		return self.settings.get("scriptFile")

	@property
	def script(self):
		return "{}.{}".format(self.func_name, self.scriptFile[:-3])
	
	def load_main(self):
		return getattr(import_module(self.script), "main")



import sys
from typing import Callable
from .settings import load_project
from .functions import Function, Project
from . import logger


ROUTE_INFO = tuple[str, list[str], Callable, str]


def parse_project(path: str) -> list[ROUTE_INFO]:
	r = load_project(path)
	sys.path.insert(1, path)
	p = Project(r["project"])
	params = []
	for f in r["functions"]:
		try:
			t = Function(*f)
			url = "{}/{}".format(p.route_prefix, t.trigger.route)
			logger.info("[{}] {} {}".format(t.func_name, ", ".join(t.trigger.methods).upper(), url))
			params.append((
				url,
				t.trigger.methods,
				t.load_main(),
				t.trigger.name
			))
		except ValueError as e:
			logger.warning("[{}] Unable to mount Function: {}".format(f[0], e))
	return params



from setuptools import setup, find_packages
from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


NAME = "pycactus"
VERSION = "0.1.0"
REQUIRES = ["azure-functions"]
FLASK_REQUIRES = ["flask"]


setup(
    name=NAME,
    version=VERSION,
    description="Adapter to run an Azure Function Application with a WSGI Web Server.",
    author="Claudjos",
    author_email="claudjosmail@gmail.com",
    url="https://github.com/Claudjos/cactus",
    keywords=["Azure Function", "Web App"],
    install_requires=REQUIRES,
    extras_require={
        "flask": FLASK_REQUIRES
    },
    packages=find_packages(),
    package_data={},
    include_package_data=False,
    long_description_content_type='text/markdown',
    long_description=long_description
)



import flask, inspect, re
import azure.functions as func
from typing import Callable
from .route_info import parse_project, ROUTE_INFO
from . import logger


FIND_PARAMS_NAMES_REGEX = re.compile(r"\{(.*?)\}")
FIND_REGEX_CONSTRAINT = re.compile(r"regex\(([^}]*)\)")


def adjust_route_for_flask(route: str) -> str:
	"""
	Converts the route definition of Azure Function to the Flask format.
	The only constraint supported is "regex:()".
	"""
	for param in FIND_PARAMS_NAMES_REGEX.findall(route):
		if ":" in param:
			name, constraint = param.split(":", 1)
			try:
				constraint = FIND_REGEX_CONSTRAINT.findall(constraint)[0]
				route = route.replace("{"+param+"}",
					"<regex(\"{}\"):{}>".format(constraint, name))
			except:
				logger.warning("Constraint not supported: {}. Fall back to route".format(tup[0]))
				route = route.replace("{"+param+"}", "<string:{}>".format(param))
		else:
			route = route.replace("{"+param+"}", "<string:{}>".format(param))

	return "{}".format(route)


def flask_request_to_azure(req: flask.Request) -> func.HttpRequest:
	return func.HttpRequest(
		req.method,
		req.path,
		headers={**req.headers},
		route_params=req.view_args,
		params={**req.args},
		body=req.stream.read()
	)


def azure_response_to_flask(res: func.HttpResponse) -> flask.Response:
	return flask.Response(
		res.get_body(),
		status=res.status_code,
		headers={**res.headers}
	)


def wrap_handler(handler: Callable):
	def wrapper(*args, **kwargs):
		from flask import request
		return azure_response_to_flask(handler(flask_request_to_azure(request)))
	wrapper.__name__ = "{}_{}".format(
		handler.__module__.replace(".", "_"),
		handler.__name__
	)
	return wrapper


def build_blueprint(name: str, route_infos: list[ROUTE_INFO]) -> flask.Blueprint:
	blueprint = flask.Blueprint(name, __name__)
	for path, methods, handler, input_name in route_infos:
		logger.debug("[{}] {}".format(",".join(methods), path))
		blueprint.route(
			adjust_route_for_flask(path),
			methods=methods,
			strict_slashes=False
		)(wrap_handler(handler))
	return blueprint


def build_app(path: str) -> flask.app.Flask:
	app = flask.Flask(__name__)
	app.register_blueprint(
		build_blueprint(
			f"FunctionApp <{path}>", 
			parse_project(path)
		)
	)
	return app
